///////////////////////////// Constants ////////////////////////////////////////////

// Number of used nets
const uint NET_COUNT = 3;

// When set to true then always the original inputs are used for combination with the outputs of the 
// last processed net. Else the previously (already combined) inputs are used for the next combination
const bool USE_ORIGINAL_INPUTS_FOR_COMBINE = false;
                                                   
// Prefix for all net file names. A number starting with 1 is appended to this prefix for loading the nets
const string NET_FILE_NAME_PREFIX = "Net_";

// Name of the input file
const string INPUT_FILE_NAME = "Input.csv";

// Name of the output file
const string RESULT_FILE_NAME = "Results.csv";




///////////////////////////// Static variables ////////////////////////////////////////////

// The currently processed net number. Valid values start from 1.
uint sNetNumber = 0;

// Holding the number of inputs and outputs of the net (must be the same for this script to work)
uint sIOCount = 0;


// Script entry point
void main()
{
    Init();
    
    bool ok = true;
    
    while (ok && (sNetNumber < NET_COUNT)) 
    {
        ok = ProcessNextNet();
    } 
    
    if (!ok)
    {
        MessageBox("Unable to process net # " + sNetNumber + "." + "\r\n\r\n" +
                   "This is most probably related to an incompatible number of inputs and outputs.\r\n" + 
                   "For this script to work all nets must have the same number of inputs and outputs and " +
                   "the number of inputs must be the same as the number of outputs."); 
    }
    
    // Export the results to file
    ExportResults();
    
    CleanUp();
}


// One-time initialization
void Init()
{
    // Adjust the view for optimum performance
    ViewSetting(SHOW_LINKS, false);
    ViewSetting(UPDATE_THINK, false);
    ViewSetting(SHOW_ACT_SPIKES, false);
    ViewSetting(SHOW_FIRE, false);
}


// Perform some clean-up after script execution
void CleanUp()
{
    // Ensure that recording is stopped
    StopRecording();
    
    // Restore view to most common options
    ViewSetting(SHOW_LINKS, true);
    ViewSetting(UPDATE_THINK, true);
    EnableLessonEditorUpdate(true);
}

// Process the next neural net
bool ProcessNextNet()
{
    bool ok = true;
    
    // Adjust for next net number
    sNetNumber++;
    
    // Build net name and try to open the net
    string scriptName = NET_FILE_NAME_PREFIX + sNetNumber + ".mbn";
    
    OpenNet(scriptName);
    
    if (sNetNumber == 1)
    { // Processing the first net
        // This script will only work if the nets have an equal number of inputs and outputs 
        sIOCount = GetInputCount();
        ok = (sIOCount >= 1) && (sIOCount == GetOutputCount());
        
        if (ok)
        {
            // Ensure that lesson editor update is enabled when input data is loaded (processing first net).
            EnableLessonEditorUpdate(true);
            // Ensure there are two lessons
            SetLessonCount(2);
            
            // Ensure that the second lesson is in the correct format for the net and is empty
            SelectLesson(2);
            NamesFromNet();
            ClearLesson();
            
            // Import the input data into the first lesson
            SelectLesson(1);
            NamesFromNet();
            ClearLesson();
            ImportLessonInputsRaw(INPUT_FILE_NAME);
            
            // From now on when performing Think Steps all input and output data from the net will end up as
            // a new pattern in lesson #2 
            StartRecording(2);
        }
    }
    else 
    { // Not the first net
        // This script only works if all the nets have the same input and output count
        ok = (sIOCount == GetInputCount());
        ok = ok && (sIOCount == GetOutputCount());
    }
    
    if (ok)
    {
        // Disable update of lesson editor for maximum performance
        EnableLessonEditorUpdate(false);
        
        // Clear lesson #2
        SelectLesson(2);
        ClearLesson();
            
        // Select Lesson #1 and perform a Think Step on each pattern. Record inputs and outputs to lesson #2
        SelectLesson(1);
        ThinkLesson();
        SleepExec();
        
        if (sNetNumber < NET_COUNT)
        { // Last net not yet processed. We must combine the just recorded results into lesson #1
          // for the next net to be processed.
            ok = CombineLessons();
        }
    }
    
    return ok;
}

// Combine the output data recorded in lesson #2 with the input data contained in lesson #1.
// The new resulting input lesson will end up in lesson #1.
bool CombineLessons()
{ 
    bool ok = true;
    
    if (USE_ORIGINAL_INPUTS_FOR_COMBINE)
    { // Re-import the original input data before combination if this is configured.      
        SelectLesson(1);
        ImportLessonInputsRaw(INPUT_FILE_NAME);
    }
    
    uint patternCount = GetLessonSize();
    
    // For every pattern in the lesson...
    for (uint pattern = 1; pattern <= patternCount; pattern++)
    {
        double oldInput;
        double recordedOutput;
        double newInput;
        
        // For every column, i.e. every input in the lesson
        for (uint column = 1; column <= sIOCount; column++)
        { 
            // Read recorded output value from lesson #2
            SelectLesson(2);
            SelectPattern(pattern);
            ok = GetPatternOutput(column, recordedOutput); 
            
            // Read input value from lesson #1
            SelectLesson(1);
            SelectPattern(pattern);
            ok = ok && GetPatternInput(column, oldInput);         
            
            // Calculate new input value
            newInput = CombineDataSet(oldInput, recordedOutput);
            
            // Now wqrite back the new input value to lesson #1
            if (ok)
            {
                SetPatternInput(column, newInput); 
            }
            else
            { // Abort in case of an error (should never happen)
                break;
            }
        }
        
        if (!ok)
        { // Abort in case of an error (should never happen)
            break;
        }   
    }
    
    return ok; 
}

// This function defines how the input will be combined with the last recorded output
double CombineDataSet(double input, double recordedOutput)
{
    double combined = input - recordedOutput;
    
    return combined;
}

// Export the results to file
void ExportResults()
{
    // The inputs and outputs from the last net have been recorded to lesson #2.
    // We are fine with the outputs but we want to export these outputs together with 
    // the original inputs.
    // Thus, we import the original inputs of lesson #2 again from file before we export the whole 
    // lesson.
    
    SelectLesson(2);
    ImportLessonInputsRaw(INPUT_FILE_NAME);
    ExportLessonRaw(RESULT_FILE_NAME);
    
    Trace("Results exported to file " + RESULT_FILE_NAME + "\r\n"); 
}


// Automatically called when the user aborts the script
void OnAbortScript()
{
    // This ensures that cleanup is also performed when the user aborts the script
    CleanUp();
}